(* ast.ml : this file defines the Abstract Syntax Tree: the first intermediate *)
(* representation used *)

type symbol = Symbol.symbol
type pos = Error_handler.pos


(************************************)
(***** ABSTRACT SYNTAX **************)
(************************************)

type ast =
    {
      ast_name : symbol ;
      ast_const :  dec ;
      ast_ty :  dec ;
      ast_var :  dec ;
      ast_proc :  dec ;
      ast_body :  exp ;
      ast_pos : pos ;
    }
and  exp =
  | Var_exp of  var
  | Null_exp
  | Int_exp of int * pos
  | Const_exp of symbol * pos
  | String_exp of string * pos
  | Bool_exp of bool * pos
  | Call_exp of  call_exp
  | Op_exp of  op_exp
  | Seq_exp of  seq_exp
  | Assign_exp of  assign_exp
  | If_exp of  if_exp
  | While_exp of  while_exp
  | For_exp of  for_exp
  | Return_exp of  return_exp

and  var =
  | Exp_var of exp
  | Var_or_Const of symbol * pos
  | Field_access of  var * symbol * pos
  | Array_access of  var *  exp * pos
  | Fun_access of exp

and  dec =
    Const_dec of  const_dec list
  | Type_dec of  ty_dec list
  | Var_dec of  var_dec list
  | Proc_dec of  proc_dec list
  | None_dec


(**************** Details ************)
and  const_value_type =
    Def of  exp | Undef of (symbol * pos)

and  const_dec =
    {
      const_name : symbol ;
      const_ty : ty_const ;
      const_value :  const_value_type ;
      const_pos : pos
    }
and ty_const = Int | Bool | String | Other

and  ty_dec =
    {
      ty_name : symbol ;
      ty_ty :  ty_ty ;
      ty_pos : pos ;
    }

and  ty_ty =
| Existing_ty of symbol * pos
| Record_ty of  field list
| Array_ty of  array_ty

and  array_ty =
    {
      array_range_left :  exp ;
      array_range_right :  exp ;
      array_elts_ty :  ty_ty ;
      array_pos : pos ;
    }

and  field =
    {
      field_names : (symbol * pos)  list ;
      field_ty :  ty_ty ;
      field_pos : pos
    }

and  var_dec =
    {
      var_dec_names : (symbol*pos)  list ;
      var_dec_escapes : bool ref list ;
      var_dec_ty :  ty_ty ;
      var_dec_pos : pos ;
    }

and  proc_dec =
    {
      proc_name : symbol ;
      proc_ret_ty : symbol ;
      mutable proc_depth : int ;
      proc_params :  param list ;
      proc_const :  dec ;
      proc_ty_dec :  dec ;
      proc_var_dec :  dec ;
      proc_proc_dec :  dec ;
      proc_body :  exp ;
      proc_pos : pos ;
    }

and  param =
    {
      param_name : symbol ;
      param_mode : bool ;
      param_ty : symbol ;
      param_pos : pos ;
    }


and  call_exp =
    {
      call_proc : symbol ;
      call_args :  exp list ;
      call_pos : pos
    }

and  op_exp =
    Un of  un_exp
  | Bin of  bin_exp

and  un_exp =
    {
      un_oper :  exp ; (* operand *)
      un_op : op ;
      un_pos : pos
    }

and  bin_exp =
    {
      bin_left_oper :  exp ;
      bin_right_oper :  exp ;
      bin_op : op ;
      bin_pos : pos
    }

and op = Less | Great | Equals | Leq | Geq | Or | And | Add | Sub | Mul | Div | Mod | Not | Diff

and  seq_exp =  exp list

and  assign_exp =
    {
      assign_var :  var ;
      assign_exp :  exp ;
      assign_pos : pos
    }

and  if_exp =
    {
      if_test :  exp ;
      if_then :  exp ;
      if_else :  exp option ;
      if_pos : pos ;
    }

and  while_exp =
    {
      while_test :  exp ;
      while_body :  exp ;
      while_pos : pos ;
    }

and  for_exp =
    {
     for_var :  var ;
     for_start :  exp ;
     for_end :  exp ;
     for_to : bool ;             (* true = TO ; false = DOWNTO *)
     for_body :  exp ;
     for_pos : pos ;
    }

and  return_exp =
    {
      return_exp :  exp ;
      return_pos : pos ;
    }


(* [indent i] creates a space of size [i] *)
let rec indent = function
  | 0 -> ""
  | i -> "  " ^ indent (i-1)


(* convert an ast into a string *)
  let rec ast_s d ast =
    indent d ^ "AST " ^ Symbol.string ast.ast_name ^ "\n" ^
    dec_s (d+1) ast.ast_const ^
    dec_s (d+1) ast.ast_ty ^
    dec_s (d+1) ast.ast_var ^
    dec_s (d+1) ast.ast_proc ^
    indent d ^ "body :\n" ^ exp_s (d+1) ast.ast_body

    and dec_s d = function

      | Const_dec l ->
	  indent d ^
	  "Constant(s) dec :\n" ^
	  List.fold_left
	    (
	      fun s e -> s ^ const_dec_s (d+1) e ^"\n"
	    ) "" l

    | Type_dec l ->
	indent d ^
	"Type(s) dec :\n" ^
	List.fold_left
	  (
	    fun s e -> s ^ type_dec_s (d+1) e ^"\n"
	  ) "" l

      | Var_dec l ->
	  indent d ^
	  "Variable(s) dec :\n" ^
	  List.fold_left
	    (
	      fun s e -> s ^ var_dec_s (d+1) e ^"\n"
	    ) "" l

      | Proc_dec l ->
	  indent d ^
	  "Procedures(s) dec :\n" ^
	  List.fold_left
	    (
	      fun s e -> s ^ proc_dec_s (d+1) e ^"\n"
	    ) "" l

      | None_dec -> ""

    and const_dec_s d e =
      indent d ^ "name : " ^
      Symbol.string e.const_name ^
      ", type : " ^
      (
	match e.const_ty with
	   Int -> "int"
	 | Bool -> "bool"
	 | String -> "string"
	 | Other -> "undefined") ^
      "\n" ^
      indent d ^
      "value :\n" ^
      (
	match e.const_value with
	   Undef (a,_) -> indent (d+1) ^ Symbol.string a
	 | Def v -> exp_s (d+1) v
      )

    and type_dec_s d t =
      indent d ^
      "name : " ^
      Symbol.string t.ty_name ^
      "\n" ^
      ty_ty_s d t.ty_ty

    and ty_ty_s d e =
	indent d ^ (
	  match e with

	    | Record_ty l ->
		"record :\n" ^
		List.fold_left
		  (
		    fun s e -> s ^ field_s (d+1) e ^ "\n"
		  ) "" l

	    | Array_ty a ->
		"array :\n" ^
		indent (d+1) ^
		"from\n" ^
		exp_s (d+1) a.array_range_left ^
		"\n" ^
		indent (d+1) ^
		"to\n" ^
		exp_s (d+1) a.array_range_right ^
		"\n" ^
		indent (d+1) ^
		"of type :" ^
		ty_ty_s (d+1) a.array_elts_ty ^
		"\n"
	    | Existing_ty (a,_) -> indent d ^ Symbol.string a

	)


    and field_s d f =
      indent d ^
      "names : " ^
      (
	List.fold_left
	 (
	   fun s e ->
	     s ^
	     Symbol.string (fst e)^
	     " "
	 ) "" f.field_names
      ) ^
      " and type : " ^
      ty_ty_s (d+1) f.field_ty


    and var_dec_s d v =
      indent d ^
      "names : " ^
      (
	List.fold_left
	 (
	   fun s (e,esc) ->
	     s ^
	     (if !esc then "@" else "") ^
	     Symbol.string (fst e)
	     ^ " "
	 ) ""
	 ( List.combine v.var_dec_names v.var_dec_escapes)
      ) ^
      " and type : " ^
      ty_ty_s (d+1) v.var_dec_ty

    and proc_dec_s d p =
      indent d ^
      "procedure with name " ^
      Symbol.string p.proc_name ^
      "\n" ^
      indent (d+1) ^
      "depth " ^
      string_of_int p.proc_depth ^
      "\n" ^
      indent (d+1) ^
      "with parameters :\n" ^
      List.fold_left (fun s e -> s ^ param_s (d+1) e ^ "\n") "" p.proc_params ^
      dec_s (d+1) p.proc_const ^
      dec_s (d+1) p.proc_ty_dec ^
      dec_s (d+1) p.proc_var_dec ^
      dec_s (d+1) p.proc_proc_dec ^
      indent d ^ "body :\n" ^
      exp_s (d+1) p.proc_body

    and param_s d e =
      indent d ^ "name : " ^
      Symbol.string e.param_name ^
      "," ^ "mode : " ^
      (match e.param_mode with
	   true -> "reference"
	 | false -> "value") ^
      ", type : " ^
      Symbol.string e.param_ty

    and exp_s d = function
      | Var_exp v -> indent d ^ "variable\n" ^ var_s (d+1) v
      |	Null_exp -> indent d ^ "Null exp"
      | String_exp (s,_) -> indent d ^ "\"" ^ s ^ "\""
      | Int_exp (i,_) -> indent d ^ "integer : " ^ string_of_int i
      | Const_exp (a,_) -> indent d ^ "constant : " ^ Symbol.string a
      | Bool_exp (b,_) -> indent d ^ "boolean : " ^ string_of_bool b
      | Call_exp c -> indent d ^ "call :\n" ^ call_exp_s (d+1) c
      | Op_exp op -> indent d ^ "operation :\n" ^ op_exp_s (d+1) op
      | Seq_exp s -> indent d ^ "sequence :\n" ^ seq_exp_s (d+1) s
      | Assign_exp ass -> indent d ^ "assignation :\n" ^ assign_exp_s (d+1) ass
      | If_exp ifexp -> indent d ^ "If :\n" ^ if_exp_s (d+1) ifexp
      | While_exp wh -> indent d ^ "While :\n" ^ while_exp_s (d+1) wh
      | For_exp f -> indent d ^ "For :\n" ^ for_exp_s (d+1) f
      |	Return_exp ret -> indent d ^ "Return :\n" ^ return_exp_s (d+1) ret

    and var_s d = function
      |	Exp_var e ->  indent d ^ "var access of\n" ^ (exp_s (d+1) e)
      | Var_or_Const (a,_) -> indent d ^ "var access : " ^ Symbol.string a
      | Field_access (v,a,_) ->
	  indent d ^ "field_access :\n " ^
	  var_s d v ^ "\n" ^ indent d ^ Symbol.string a
      | Array_access (v,e,_) ->
	  indent d ^ "array access :\n" ^
	  var_s d v ^ "\n" ^ indent d ^ exp_s (d+1) e
      | Fun_access c -> indent d ^ "function access :\n" ^
	  exp_s (d+1) c

    and call_exp_s d c =
      indent d ^ "procedure called : " ^ Symbol.string c.call_proc ^ "\n" ^
      indent d ^ "parameters :" ^
      List.fold_left (fun s e -> s ^ "\n" ^ exp_s (d+1) e) "" c.call_args

    and op_exp_s d = function
      | Un e -> indent d ^ "unary operation :\n" ^ un_s (d+1) e
      | Bin e -> indent d ^ "binary operation :\n" ^ bin_s (d+1) e

    and un_s d e =
      indent d ^ "operator : " ^ op_s e.un_op ^ "\n" ^
      indent d ^ "operand :\n" ^ exp_s (d+1) e.un_oper

    and bin_s d e =
      indent d ^ "left operand :\n" ^ exp_s (d+1) e.bin_left_oper ^ "\n" ^
      indent d ^ "operator : " ^ op_s e.bin_op ^ "\n" ^
      indent d ^ "right operand :\n" ^ exp_s (d+1) e.bin_right_oper

    and op_s = function
      | Less -> "-"
      | Great -> ">"
      | Equals -> "="
      | Leq -> "<="
      | Geq -> ">="
      | Or -> "or"
      | And -> "and"
      | Add -> "+"
      | Sub -> "-"
      | Mul -> "*"
      | Div -> "/"
      | Mod -> "mod"
      | Not -> "not"
      | Diff -> "<>"

    and seq_exp_s d =
      List.fold_left
	(
	  fun s e -> s ^ "\n" ^ indent d ^ "and\n" ^ exp_s (d+1) e
	) ""

    and assign_exp_s d ass =
      indent d ^ "variable :\n" ^
      var_s (d+1) ass.assign_var ^ "\n" ^
      indent d ^ "expression assigned :\n" ^
      exp_s (d+1) ass.assign_exp

    and if_exp_s d iff =
      indent d ^ "condition :\n" ^ exp_s (d+1) iff.if_test ^ "\n" ^
      indent d ^ "positive case :\n" ^ exp_s (d+1) iff.if_then ^ "\n" ^
      (match iff.if_else with
	   None -> ""
	 | Some e -> indent d ^ "positive case :\n" ^ exp_s (d+1) e  )

    and while_exp_s d wh =
      indent d ^ "condition :\n" ^ exp_s (d+1) wh.while_test ^ "\n" ^
      indent d ^ "body :\n" ^ exp_s (d+1) wh.while_body

    and for_exp_s d f =
    indent d ^ "var : \n" ^ var_s (d+1) f.for_var ^ "\n" ^
    indent d ^ "from : \n" ^ exp_s (d+1) f.for_start ^ "\n" ^
    indent d ^ "to : \n" ^ exp_s (d+1) f.for_end ^ "\n" ^
    indent d ^ "do : \n" ^ exp_s (d+1) f.for_body ^ "\n"

    and return_exp_s d ret =
      indent d ^ "expression :\n" ^ exp_s (d+1) ret.return_exp ^ "\n"


let to_string =
  fun a -> (ast_s 0 a ^ "\n")



  open Printf

  (* convert a type into a string without indentation *)
  let rec ty_ty_to_string e =
    match e with

      | Record_ty l ->
	  "record :\n" ^
	  List.fold_left
	    (
	      fun s e -> s ^ field_to_string e ^ "\n"
	    ) "" l

	    | Array_ty a ->
		"array " ^
		"from " ^
		exp_s 0 a.array_range_left ^
		"\n" ^
		"to " ^
		exp_s 0  a.array_range_right ^
		"\n" ^
		"of type :" ^
		ty_ty_to_string a.array_elts_ty ^
		"\n"
	    | Existing_ty (a,_) -> Symbol.string a

    and field_to_string f =
      "names : " ^
      (
	List.fold_left
	 (
	   fun s e ->
	     s ^
	     Symbol.string (fst e)^
	     " "
	 ) "" f.field_names
      ) ^
   " and type : " ^
   ty_ty_to_string f.field_ty


  exception None_dec_exn


  let to_graph file ast =

    Printf.eprintf "Printing AST in \"%s\"..." file ;

    let fd = open_out_bin file in
    let nodes = ref 0 in
      fprintf fd "digraph g {\n\tordering = out ;\n" ;

      let node lab =
	incr nodes ;
	fprintf fd "\tnode%d[label = \"%s\"];\n" !nodes lab ;
	!nodes
    in

    let link x y =
      fprintf fd "\tnode%d -> node%d;\n" x y
    in

    let rec dotty_ast ast =
      let prg = node ( "Program \\n" ^ ast.ast_name ) in
	begin
	  try
	    let const = dotty_dec ast.ast_const in
	      link prg const ;
	  with None_dec_exn -> ()
	end;
	begin
	  try
	    let types = dotty_dec ast.ast_ty in
	      link prg types ;
	  with None_dec_exn -> ()
	end;
	begin
	  try
	    let proc = dotty_dec ast.ast_proc in
	      link prg proc ;
	  with None_dec_exn -> ()
	end;
	begin
	  try
	    let var = dotty_dec ast.ast_var in
	      link prg var ;
	  with None_dec_exn -> ()
	end;
	let body = dotty_exp ast.ast_body in
	let body_node = node "Body" in
	  begin
	    link body_node body ;
	    link prg body_node ;
	  end

    and dotty_dec = function
      | Const_dec l ->
	  let const_dec = node "Constants" in
	    List.iter
	      (
		fun x ->
		  let n_c =
		    node
		      (
			Symbol.string x.const_name
		      )
		  in
		  let n_t =
		    node
		      (
			"type \\n"^
			(match x.const_ty with
			   | Int -> "integer"
			   | Bool -> "boolean"
			   | String -> "string"
			   | _ -> "unknown"
			)
		      )
		  in
		  let n_v =
		    node "value \\n"
		  in
		  let n_val =
		      match x.const_value with
			  Def e ->
			    dotty_exp e
			| Undef (s,_) ->
			    node (Symbol.string s)
		  in
		    begin
		      link n_c n_t ;
		      link n_c n_v ;
		      link n_v n_val ;
		      link const_dec n_c ;
		    end
	      ) l ;
 	    const_dec

      | Type_dec l ->
	  let ty_dec = node "Types"
	  in
	    List.iter (
	      fun x ->
		let ty_n = node ("name\\n" ^
				 Symbol.string x.ty_name)
		in
		let ty_ty_n =
		  dotty_ty_ty x.ty_ty
		in
		  link ty_dec ty_n ;
		  link ty_n ty_ty_n
	    ) l ;
	    ty_dec

      | Var_dec l ->
	  let var_dec = node "Variables"
	  in
	    List.iter (
	      fun x ->
		  List.iter
		    (
		      fun (v,e) ->
			let n =
			  node
			    (
			      Symbol.string (fst v) ^
			      ( if !e then
				  " (esc) :\\n"
				else ":\\n") ^
			      String.escaped (ty_ty_to_string x.var_dec_ty)
			    )
			in
			  link var_dec n
		    )
		    (
		      List.combine
		       x.var_dec_names
		       x.var_dec_escapes
		    )
	    ) l ;
	    var_dec

      | Proc_dec l ->
	  let node_proc = node "Procedures"
	  in
	    List.iter (
	      fun p ->
		let prg = node (
		  Symbol.string p.proc_name ^ "\\n" ^
		  "depth: " ^ string_of_int p.proc_depth)
		in
		let param prc =
		  match prc.proc_params with
		      [] -> raise None_dec_exn
		    | l ->
			let np = node "arguments"
			in
			  List.iter
			    ( fun a ->
				let npa =
				  node (
				    Symbol.string a.param_name ^
				    (if a.param_mode then
				       "(esc):"
				     else
				       ":")^
				    Symbol.string a.param_ty
				  )
				in
				  link np npa
			    ) l ;
			  np
		in
		  begin
		  try
		    let n_p = param p in
		      link prg n_p ;
		  with
		      None_dec_exn -> ()
		  end ;
		  begin
			try
			  let n_c = dotty_dec p.proc_const in
			    link prg n_c ;
			with None_dec_exn -> ()
		  end ;
		  begin
		    try
		      let n_ty = dotty_dec p.proc_ty_dec in
			link prg n_ty ;
		    with None_dec_exn -> ()
		  end ;
		  begin
		    try
		      let n_var = dotty_dec p.proc_var_dec in
			link prg n_var ;
		    with None_dec_exn -> ()
		  end;
		  begin
		    try
		      let n_proc = dotty_dec p.proc_proc_dec in
			link prg n_proc ;
		    with None_dec_exn -> ()
		  end;
		  let n_bod = dotty_exp p.proc_body
		  in
		  let n_body = node "body"
		  in
		    begin
		      link prg n_body ;
		      link n_body n_bod ;
		      link node_proc prg ;
		    end
	    ) l ;
	    node_proc

      | None_dec -> raise None_dec_exn

    and dotty_ty_ty = function
	Record_ty lf ->
	  let ty = node "record"
	  in
	    List.iter (
	      fun x ->
		let nf =
		  node (
		    (
		      List.fold_left
		       (
			 fun r -> fun s ->
					     r ^
			   Symbol.string (fst s)^
			   " "
		       ) "" x.field_names
		    ) ^
		    " : \\n" ^

 		    String.escaped (ty_ty_to_string x.field_ty)
		  )
		in
		  link ty nf
	    ) lf ; ty

      | Array_ty a ->
	  let ty = node "array"
	  in
	  let nl =
	    dotty_exp a.array_range_left
			    in
			    let nr =
			      dotty_exp a.array_range_right
			    in
			    let nty =
			      node (String.escaped (ty_ty_to_string a.array_elts_ty))
			    in
			      begin
				link ty nl ;
				link ty nr ;
				link ty nty ;
			      end
			      ;
			      ty
      |	_ -> 0 (* TODO *)
    and dotty_exp = function
      | Var_exp v ->
	  dotty_var v

      |	Null_exp ->
	  node "null exp"

      | Int_exp (i,_) ->
	  node ("int " ^ string_of_int i)

      | Const_exp (a,_) ->
	  node (Symbol.string a)

      | String_exp (s,_) ->
	  node ("\'" ^ s ^ "\'")

      | Bool_exp (b,_) ->
	  node ("bool " ^ string_of_bool b)

      | Call_exp c ->
	  dotty_call c

      | Op_exp op ->
	  dotty_op op

      | Seq_exp s ->
	  let n = node "sequence"
	  in
	    List.iter
	      (
		fun e ->
		  link n
		  (dotty_exp e)
	      ) s
	    ; n

      | Assign_exp ass ->
	  dotty_ass ass
      | If_exp ifexp ->
	  dotty_if ifexp

      | While_exp wh ->
	  dotty_while wh

      | For_exp f ->
	  dotty_for f

      |	Return_exp ret ->
	  dotty_return ret

    and dotty_var = function
      |	Exp_var e ->
	  let n = node ("exp")
	  and ne = dotty_exp e in
	  link n ne ;
	  n
      | Var_or_Const (a,_) -> node (Symbol.string a)

      | Field_access (v,a,_) ->
	  let n = node ("record")
	  in
	  let nv = dotty_var v
	  in
	  let nf = node (Symbol.string a)
	  in
	    begin
	      link n nv ;
	      link n nf ;
	      n
	    end

      | Array_access (v,e,_) ->
	  let n = node ("array")
	  in
	  let nv = dotty_var v
	  in
	  let ni = dotty_exp e
	  in
	    begin
	      link n nv ;
	      link n ni ;
	      n
	    end

      | Fun_access c ->
	  let n = node ("function access")
	  in
	  let ne = dotty_exp c
	  in
	    begin
	      link n ne ;
	      n
	    end

    and dotty_call c =

      let n = node "call"
      in
      let nn = node (Symbol.string c.call_proc)
      in
	link n nn ;
	List.iter
	  (fun e->
	     let ne = dotty_exp e
	     in
	       link n ne
	  ) c.call_args ;
	n



    and dotty_op = function
      | Un e -> dotty_un e
      | Bin e -> dotty_bin e

    and op_s = function
      | Less -> "<"
      | Great -> ">"
      | Equals -> "="
      | Leq -> "<="
      | Geq -> ">="
      | Or -> "or"
      | And -> "and"
      | Add -> "+"
      | Sub -> "-"
      | Mul -> "*"
      | Div -> "/"
      | Mod -> "mod"
      | Not -> "not"
      | Diff -> "<>"

    and dotty_un e =
      let n = node (op_s e.un_op)
      in
      let exp = dotty_exp e.un_oper
      in
	begin
	  link n exp ;
	  n ;
	end

    and dotty_bin e =
      let n = node (op_s e.bin_op)
      in
      let expl = dotty_exp e.bin_left_oper
      in
      let expr = dotty_exp e.bin_right_oper
      in
	begin
	  link n expl ;
	  link n expr ;
	  n ;
	end

    and dotty_ass ass =

      let n = node ":="
      in
      let nl = dotty_var ass.assign_var
      in
      let nr = dotty_exp ass.assign_exp
      in
	begin
	  link n nl ;
	  link n nr ;
	  n ;
	end

    and dotty_if i =

      let n = node "if"
      in
      let t = dotty_exp i.if_test
      in
	link n t;
      let th = dotty_exp i.if_then
      in
	link n th ;
	begin
	  match i.if_else with
	      None -> ()
	    | Some e ->
		let el = dotty_exp e
		in
		  link n el ;
	end;
	n

    and dotty_while wh =
      let n = node "while"
      in
      let t = dotty_exp wh.while_test
      in
	link n t;
	let th = dotty_exp wh.while_body
	in
	  link n th ;
	  n

    and dotty_for f =
      let n = node "for" in
      let nv = dotty_var f.for_var in
	link n nv;
	let n_from = dotty_exp f.for_start in
	  link n n_from;
	  let n_to = dotty_exp f.for_end in
	    link n n_to;
	    let n_bod = dotty_exp f.for_body in
	      link n n_bod;
	      n

    and dotty_return ret =
      let n = node "return"
      in
      let t = dotty_exp ret.return_exp
      in
	link n t;
	  n

    in
      ignore( dotty_ast ast ) ;

      fprintf fd "}\n" ;
      close_out fd ;
      Printf.eprintf "ok\n" 



  let ast_to_proc ast =
    {
      proc_name = "main" ;
      proc_ret_ty = "void" ;
      proc_depth = 0 ;
      proc_params = [] ;
      proc_const = ast.ast_const ;
      proc_ty_dec = ast.ast_ty ;
      proc_var_dec = ast.ast_var ;
      proc_proc_dec = ast.ast_proc ;
      proc_body = ast.ast_body ;
      proc_pos = ast.ast_pos ;
    }


  exception Position_unknown

  let fail () = raise Position_unknown

  let rec get_pos_exp e = match e with

    | Var_exp v -> get_pos_var v
    | Null_exp -> fail ()
    | Int_exp _-> fail ()
    | Const_exp (_,p) -> p
    | String_exp _ -> fail ()
    | Bool_exp _ -> fail ()
    | Call_exp c -> c.call_pos
    | Op_exp op -> get_pos_op op
    | Seq_exp l -> get_pos_exp (List.hd l)
    | Assign_exp ass -> ass.assign_pos
    | If_exp iff -> iff.if_pos
    | While_exp wh -> wh.while_pos
    | For_exp f -> f.for_pos
    | Return_exp ret -> ret.return_pos

  and get_pos_var = function

      Var_or_Const (_,p) -> p
    | Field_access (_,_,p) -> p
    | Array_access (_,_,p) -> p
    | Fun_access (Call_exp c) -> c.call_pos
    | _ -> failwith "other case not expected in Ast.get_pos_var"

  and get_pos_op = function

      Un un -> un.un_pos
    | Bin bin -> bin.bin_pos


