type symbol = string

type pos = Error_handler.pos
(** position of the lexeme *)

(************************************)
(***** ABSTRACT SYNTAX **************)
(************************************)
(** abstract syntax *)
type ast =
    {
      ast_name : symbol ; (* program name *)
      ast_const :  dec ;  (* constant declarations *)
      ast_ty :  dec ;     (* types declarations *)
      ast_var :  dec ;    (* variables declation *)
      ast_proc :  dec ;   (* procedures declaration *)
      ast_body :  exp ;
      ast_pos : pos ;     (* position in input file *)
    }
    (* Expressions *)
and exp =
  | Var_exp of  var   (* variable *)
  | Null_exp          (* void expression, e.g. returned by procedures *)
  | Int_exp of int*pos           (* integer *)
  | Const_exp of symbol * pos    (* constant *)
  | String_exp of string * pos   (* string *)
  | Bool_exp of bool * pos       (* boolean *)
  | Call_exp of  call_exp        (* procedure call *)
  | Op_exp of  op_exp            (* operation *)
  | Seq_exp of  seq_exp          (* sequence of instructions *)
  | Assign_exp of  assign_exp    (* assignment x := y *)
  | If_exp of  if_exp            (* if block *)
  | While_exp of  while_exp      (* while loop *)
  | For_exp of  for_exp          (* for loop *)
  | Return_exp of  return_exp    (* exit procedure *)

    (* Variables *)
and  var =
  | Exp_var of exp
  | Var_or_Const of symbol * pos
  | Field_access of  var * symbol * pos
  | Array_access of  var *  exp * pos
  | Fun_access of exp

    (* Declarations *)
and  dec =
    Const_dec of  const_dec list
  | Type_dec of  ty_dec list
  | Var_dec of  var_dec list
  | Proc_dec of  proc_dec list
  | None_dec


(**************** Details ************)
and  const_value_type =
    Def of  exp | Undef of (symbol * pos)

and  const_dec =
    {
      const_name : symbol ;
      const_ty : ty_const ;
      const_value :  const_value_type ;
      const_pos : pos
    }

and ty_const = Int | Bool | String | Other

and  ty_dec =
    {
      ty_name : symbol ;
      ty_ty :  ty_ty ;
      ty_pos : pos ;
    }

    (* language types *)
and  ty_ty =
| Existing_ty of symbol * pos  (* redefinition of a type *)
| Record_ty of  field list     (* record type *)
| Array_ty of  array_ty        (* array type *)

and  array_ty =
    {
      array_range_left :  exp ;
      array_range_right :  exp ;
      array_elts_ty :  ty_ty ;
      array_pos : pos ;
    }

and  field =
    {
      field_names : (symbol  * pos) list ;
      field_ty :  ty_ty ;
      field_pos : pos
    }

and  var_dec =
    {
      var_dec_names : (symbol * pos)  list ;
      var_dec_escapes : bool ref list ;
      var_dec_ty :  ty_ty ;
      var_dec_pos : pos ;
    }

and  proc_dec =
    {
      proc_name : symbol ;
      proc_ret_ty : symbol ;      (* return type *)
      mutable proc_depth : int ;
      proc_params :  param list ;
      proc_const :  dec ;
      proc_ty_dec :  dec ;
      proc_var_dec :  dec ;
      proc_proc_dec :  dec ;
      proc_body :  exp ;
      proc_pos : pos ;
    }

and  param =
    {
      param_name : symbol ;
      param_mode : bool ;    (* escaped or not ? (keyword "var") *)
      param_ty : symbol ;
      param_pos : pos ;
    }


and  call_exp =
    {
      call_proc : symbol ;
      call_args :  exp list ;
      call_pos : pos
    }

and  op_exp =
    Un of  un_exp    (* unary operator *)
  | Bin of  bin_exp  (* binary operator *)

and  un_exp =
    {
      un_oper :  exp ; (* operand *)
      un_op : op ;
      un_pos : pos
    }

and  bin_exp =
    {
      bin_left_oper :  exp ;
      bin_right_oper :  exp ;
      bin_op : op ;
      bin_pos : pos
    }

and op = Less | Great | Equals
	 | Leq | Geq | Or | And
	 | Add | Sub | Mul
	 | Div | Mod | Not | Diff

and  seq_exp =  exp list

and  assign_exp =
    {
      assign_var :  var ;
      assign_exp :  exp ;
      assign_pos : pos
    }

and  if_exp =
    {
      if_test :  exp ;
      if_then :  exp ;
      if_else :  exp option ;
      if_pos : pos ;
    }

and  while_exp =
    {
      while_test :  exp ;
      while_body :  exp ;
      while_pos : pos ;
    }

and  for_exp =
    {
     for_var :  var ;
     for_start :  exp ;
     for_end :  exp ;
     for_to : bool ;      (* true = TO ; false = DOWNTO *)
     for_body :  exp ;
     for_pos : pos ;
    }

and  return_exp =
    {
      return_exp :  exp ;
      return_pos : pos ;
    }

val to_string : ast -> string

val to_graph : string -> ast -> unit
(** [to_graph filename ast] creates a file named [filename]
    which contains the dotty representation of [ast] *)

val ast_to_proc : ast -> proc_dec
(** convert an ast into a procedure named "main" *)

exception Position_unknown

val get_pos_exp : exp -> pos
(** [get_pos_exp exp] return the position in source code
  of [exp]. Raise [Position_unknown] if not found. *)
